==> lmtpd-0.9.19.tar.gz: lmtpd-0.9.19/checkrelay.c <==
/*
   Copyright 2000, 2001, 2002, 2003 Laurent Wacrenier

   This file is part of PLL lmtpd.

   PLL lmtpd is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   PLL lmtpd is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with PLL lmtpd; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

static char const rcsid[] =
"$Id: checkrelay.c,v 1.24 2003/12/02 12:07:38 lwa Exp $";

#include "config.h"

#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <arpa/nameser.h>
#include <arpa/inet.h>
#include <resolv.h>
#include <netdb.h>
#include <regex.h>
#include <syslog.h>

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>

#include "lmtp.h"

/* missing declarations on HPUX */
#if !HAVE_DECL_RES_SEARCH
int res_search(const char *dname, int class, int type,
	       u_char *answer, int anslen);
#endif

#if !HAVE_DECL_DN_SKIPNAME
int dn_skipname(const u_char *comp_dn, const u_char *eom);
#endif

#if 0
#define DEBUG(a, x...) fprintf(stderr, "***DEBUG " a "\n" , ##x)
#else
#define DEBUG(x...)
#endif

#define ADDR_FORMAT "\\[(([0-9]{1,3})\\.([0-9]{1,3})\\.([0-9]{1,3})\\.([0-9]{1,3}))\\]"

extern int h_errno;

#define REPLYBUF 4096

#if !HAVE_INET_NTOP
/* fake function for hpux */
char * inet_ntop(int af, const void * src, char * dst, socklen_t size) {
  if (af == AF_INET) {
    char *a = inet_ntoa(*(struct in_addr *)src);
    *dst = 0;
    strncat(dst, a, size);
    return dst;
  } else {
    errno = EPFNOSUPPORT;
    return NULL;
  }
}
#endif

/*
 *  simple DNS resolution
 *  return 0 when not found
 *         1 when found
 *         <0 on lookup error
 */

static int dns_exists(struct fstate *state, char *name,
		      char *a, char *txt) {
  u_char reply[REPLYBUF];
  u_char *reply_end;
  HEADER *reply_header;
  u_char *ap;
  int ancount, qdcount;
  int len;
  struct user_res_ctx res_ctx;

  *a=0;
  *txt=0;

  DEBUG("DNS query %s", name);

  user_res_begin(state, &res_ctx);
  len=res_search(name, C_IN, T_ANY, reply, sizeof(reply));
  user_res_end(state, &res_ctx);

  if (len < 0) {
    switch (h_errno) {
    case HOST_NOT_FOUND:
    case NO_DATA:
      return 0;
    case TRY_AGAIN:
      return -2;
    case NO_RECOVERY:
      return -3;
    default:
      return -1;
    }
  }

  if (len > sizeof(reply))
    len = sizeof(reply);

  reply_end = reply+len;

  reply_header = (HEADER *)&reply;
  ancount = ntohs(reply_header->ancount);
  qdcount = ntohs(reply_header->qdcount);

  ap = (u_char *) &reply + sizeof(HEADER);

  while(qdcount--) {
    int l ;
    if (ap + QFIXEDSZ > reply_end)
      return -9;
    l = dn_skipname(ap, reply_end);
    if (l<0) {
      return -4;
    }
    ap += l + QFIXEDSZ;
  }

  while(ancount-- > 0) {
    int type;
    int class;
    int ttl;
    int length;
    int l;
    u_char *pos = ap;
    if ((l = dn_skipname(pos, reply_end))<0)
      return -5;
    pos += l;
    if (ap + l + RRFIXEDSZ > reply_end)
      return -8;
    GETSHORT(type, pos);
    GETSHORT(class, pos);
    GETLONG(ttl, pos);
    GETSHORT(length, pos);
    if (ap + l + RRFIXEDSZ + length > reply_end)
      return -7;
    if (class != C_IN)
      return -6;
    switch(type) {
    case T_A:
      inet_ntop(AF_INET, pos, a, ADDRESS_LEN);
      break;
    case T_TXT: {
      int txt_len = *pos++;
      int i;
      if (txt_len > TXT_RDATA_LEN) txt_len=sizeof(txt)-1;
      if (txt_len > length-1)    txt_len=length-1;
      for(i=0; i < txt_len; i++) {
	txt[i] = isprint(pos[i]) ? pos[i] : ' ';
      }
      txt[i]=0;
    }
    }
    ap += l + RRFIXEDSZ + length;
  }
  return a[0]!=0;
}

/*
 *  extract an IP adress from headers
 *
 *  returns
 *   <0   error occurs in adress format
 *   0    no address is found
 *   >0   an address is found
 */

int get_addr_from_received(char *str, int ip[4],
			   char *query, char *a, char *txt) {
  int i;
  regmatch_t pmatch[6];
  static regex_t re_addr;
  static int re_addr_ok=0;

  char *errmsg = NULL;

  if (!re_addr_ok) {
    regcomp(&re_addr, ADDR_FORMAT, REG_EXTENDED);
    re_addr_ok=1;
  }
  if (regexec(&re_addr, str, 6, pmatch, 0) != 0) {
    return 0;
  }

  for(i=1; i<=4; i++) {
    char *s=str+pmatch[i+1].rm_so;
    char *e=str+pmatch[i+1].rm_eo;
    int x=0;
    if (*s=='0' && e-s!=1) {
      if (!errmsg) errmsg="Invalid IP format";
    }
    while(s<e) {
      x=x*10+ (*s-'0');
      s++;
    }
    if (x>255) {
      if (!errmsg) errmsg= "Invalid IP address";
    }
    ip[i-1]=x;
  }

  if (ip[0]>=224) {
    if (!errmsg) errmsg = "Multicast IP address";
  }
  if (query) {
    int len = pmatch[1].rm_eo-pmatch[1].rm_so;
    strncpy(query, str+pmatch[1].rm_so, len);
    query[len]=0;
  }
  if (errmsg) {
    if (txt) strcpy(txt, errmsg);
    if (a)   strcpy(a, "127.0.0.2");
    return -1;
  } else {
    if (txt) txt[0]=0;
    if (a) a[0]=0;
  }
  return 1;
}

/* entry found */
#define BLFLAG_FOUND     0x01
/* entry not found after a full research */
#define BLFLAG_NOTFOUND  0x02
/* entry not found on a level 1 search */
#define BLFLAG_LEVEL1    0x04

/*
 * check in RHSBL blacklist type
 */

static struct blresult *check_rhsbl(struct fstate *state,
				    struct slist_elt *rbl_e ) {
  struct blresult *res=NULL;
  char *domain;
  char buf[255];
  int exists;

  DEBUG("check RHSBL"); 

  switch(rbl_e->flags) {
  case BLFLAG_FOUND:
    return (struct blresult *)rbl_e->extra;
  case BLFLAG_NOTFOUND:
    return NULL;
  }

  if (state->sender[0] == 0)
    return NULL;

  domain = strrchr(state->sender, '@');
  if (domain == NULL)
    return NULL;
  domain ++ ; /* skip @ */

  if ((res = malloc(sizeof(struct blresult)))==NULL)
    return NULL;

  snprintf(buf, sizeof(buf), "%s.%s.", domain, rbl_e->value);
  exists=dns_exists(state, buf, res->a, res->txt);
  DEBUG("exists=%d buf=%s a=%s", exists, buf, res->a);
  if (exists < 0) { /* lookup error */
    free(res);
    return NULL; 
  }
  if (exists == 0) {
    free(res);
    rbl_e->flags = BLFLAG_NOTFOUND;
    return NULL;
  }

  if (rbl_e->extra)
    free(rbl_e->extra);
  strncpy(res->query, domain, sizeof(res->query)-1);
  res->query[sizeof(res->query)-1]=0;
  rbl_e->extra = res;
  rbl_e->flags = BLFLAG_FOUND;
  res->rbl   = rbl_e->key;
  res->proof = 0;
  res->header = NULL;
  res->domain = rbl_e->value;
  return res;
}

/*
 *  check if a Received is in a RBL.
 *  return 1 if yes
 *  return 0 if no
 */
static struct blresult *check_rbl(struct fstate *state,
				  struct slist_elt *rbl_e,
				  int level1) {
  struct headers *hl;
  int proof = 0;
  int minproof = 0;
  struct blresult *res=NULL;

  DEBUG("check RBL"); 

  if (rbl_e->flags) {
    struct blresult *extra = (struct blresult *)rbl_e->extra;
    DEBUG("(cached)");
    if (level1) {
      return extra && extra->proof>1 ? NULL : extra;
    } else {
      if (extra)
	return extra;
      if (rbl_e->flags == BLFLAG_NOTFOUND)
	return NULL ;
      minproof = 1;  /* search again */
    }
  } 
  if ((res = malloc(sizeof(struct blresult)))==NULL)
      return NULL;
  rbl_e->flags = level1 ? BLFLAG_LEVEL1 : BLFLAG_NOTFOUND  ;
  
  hl=state->hdr_in_received;

  while(hl) {
    int ip[4];

    if (hl->type == HEADER_RECEIVED) {
      int invalid = 0;
      proof++;
      DEBUG("%s", hl->header);
      if (proof <= minproof) {
	DEBUG("skiped");
	hl=hl->next;
	continue;
      }
      if (rbl_e->value[0]==0) {
	int ret = get_addr_from_received(hl->header, ip, 
					 res->query, res->a, res->txt);
	DEBUG("ret=%d", ret);
	if (ret < 0) invalid = 1;
      } else {
	int ret = get_addr_from_received(hl->header, ip,
					 res->query, NULL, NULL);
	if (ret==1) {
	  char buf[255];
	  int exists;
	  snprintf(buf, sizeof(buf), "%d.%d.%d.%d.%s.",
		     ip[3], ip[2], ip[1], ip[0], rbl_e->value);
	  exists=dns_exists(state, buf, res->a, res->txt);
	  DEBUG("ret=%d exists=%d buf=%s a=%s", ret, exists, buf, res->a);
	  if (exists > 0) invalid = 1;
	}
      }
      DEBUG("Check(%s) %s => %d (%s %s)",
	    res->query, rbl_e->key, invalid, res->a, res->txt);

      if (invalid) {
	if (rbl_e->extra)
	  free(rbl_e->extra);
	rbl_e->extra = res;
	rbl_e->flags = BLFLAG_FOUND;
	res->rbl     = rbl_e->key;
	res->proof   = proof;
	res->header  = hl;
	res->domain  = rbl_e->value;
	return res;
      }
      if (level1)
	break;
    }
    hl = hl->next;
  }
  free(res);
  return NULL;
  
}

/*
 * check in blacklist 
 */


struct blresult *checkbl(struct fstate *state, char *rbl) {
  struct slist_elt *rbl_e; 
  int len;
  int level1;

  DEBUG("check in %s", rbl); 
  
  rbl_e = slist_search(&(state->rhsbl_server), rbl);
  if (rbl_e)
    return check_rhsbl(state, rbl_e);

  len = strlen(rbl);
  if (len>0 && rbl[len-1]=='*') {
    rbl[len-1]=0;
    level1=0;
  } else {
    level1=1;
  }
  rbl_e = slist_search(&(state->rbl_server), rbl);

  if (rbl_e) 
    return check_rbl(state, rbl_e, level1);

  return NULL;
}

==> lmtpd-0.10.7.tar.gz: lmtpd-0.10.7/checkrelay.c <==
/*
   Copyright 2000, 2001, 2002, 2003 Laurent Wacrenier

   This file is part of PLL lmtpd.

   PLL lmtpd is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   PLL lmtpd is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with PLL lmtpd; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "config.h"

static char const rcsid[] UNUSED =
"$Id: checkrelay.c,v 1.25 2004/10/26 09:59:19 lwa Exp $";



#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <arpa/nameser.h>
#include <arpa/inet.h>
#include <resolv.h>
#include <netdb.h>
#include <regex.h>
#include <syslog.h>

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>

#include "lmtp.h"

/* missing declarations on HPUX */
#if !HAVE_DECL_RES_SEARCH
int res_search(const char *dname, int class, int type,
	       u_char *answer, int anslen);
#endif

#if !HAVE_DECL_DN_SKIPNAME
int dn_skipname(const u_char *comp_dn, const u_char *eom);
#endif

#if 0
#define DEBUG(a, x...) fprintf(stderr, "***DEBUG " a "\n" , ##x)
#else
#define DEBUG(x...)
#endif

#define ADDR_FORMAT "\\[(([0-9]{1,3})\\.([0-9]{1,3})\\.([0-9]{1,3})\\.([0-9]{1,3}))\\]"

extern int h_errno;

#define REPLYBUF 4096

#if !HAVE_INET_NTOP
/* fake function for hpux */
char * inet_ntop(int af, const void * src, char * dst, socklen_t size) {
  if (af == AF_INET) {
    char *a = inet_ntoa(*(struct in_addr *)src);
    *dst = 0;
    strncat(dst, a, size);
    return dst;
  } else {
    errno = EPFNOSUPPORT;
    return NULL;
  }
}
#endif

/*
 *  simple DNS resolution
 *  return 0 when not found
 *         1 when found
 *         <0 on lookup error
 */

static int dns_exists(struct fstate *state, char *name,
		      char *a, char *txt) {
  u_char reply[REPLYBUF];
  u_char *reply_end;
  HEADER *reply_header;
  u_char *ap;
  int ancount, qdcount;
  int len;
  struct user_res_ctx res_ctx;

  *a=0;
  *txt=0;

  DEBUG("DNS query %s", name);

  user_res_begin(state, &res_ctx);
  len=res_search(name, C_IN, T_ANY, reply, sizeof(reply));
  user_res_end(state, &res_ctx);

  if (len < 0) {
    switch (h_errno) {
    case HOST_NOT_FOUND:
    case NO_DATA:
      return 0;
    case TRY_AGAIN:
      return -2;
    case NO_RECOVERY:
      return -3;
    default:
      return -1;
    }
  }

  if (len > sizeof(reply))
    len = sizeof(reply);

  reply_end = reply+len;

  reply_header = (HEADER *)&reply;
  ancount = ntohs(reply_header->ancount);
  qdcount = ntohs(reply_header->qdcount);

  ap = (u_char *) &reply + sizeof(HEADER);

  while(qdcount--) {
    int l ;
    if (ap + QFIXEDSZ > reply_end)
      return -9;
    l = dn_skipname(ap, reply_end);
    if (l<0) {
      return -4;
    }
    ap += l + QFIXEDSZ;
  }

  while(ancount-- > 0) {
    int type;
    int class;
    int ttl;
    int length;
    int l;
    u_char *pos = ap;
    if ((l = dn_skipname(pos, reply_end))<0)
      return -5;
    pos += l;
    if (ap + l + RRFIXEDSZ > reply_end)
      return -8;
    GETSHORT(type, pos);
    GETSHORT(class, pos);
    GETLONG(ttl, pos);
    GETSHORT(length, pos);
    if (ap + l + RRFIXEDSZ + length > reply_end)
      return -7;
    if (class != C_IN)
      return -6;
    switch(type) {
    case T_A:
      inet_ntop(AF_INET, pos, a, ADDRESS_LEN);
      break;
    case T_TXT: {
      int txt_len = *pos++;
      int i;
      if (txt_len > TXT_RDATA_LEN) txt_len=sizeof(txt)-1;
      if (txt_len > length-1)    txt_len=length-1;
      for(i=0; i < txt_len; i++) {
	txt[i] = isprint(pos[i]) ? pos[i] : ' ';
      }
      txt[i]=0;
    }
    }
    ap += l + RRFIXEDSZ + length;
  }
  return a[0]!=0;
}

/*
 *  extract an IP adress from headers
 *
 *  returns
 *   <0   error occurs in adress format
 *   0    no address is found
 *   >0   an address is found
 */

int get_addr_from_received(char *str, int ip[4],
			   char *query, char *a, char *txt) {
  int i;
  regmatch_t pmatch[6];
  static regex_t re_addr;
  static int re_addr_ok=0;

  char *errmsg = NULL;

  if (!re_addr_ok) {
    regcomp(&re_addr, ADDR_FORMAT, REG_EXTENDED);
    re_addr_ok=1;
  }
  if (regexec(&re_addr, str, 6, pmatch, 0) != 0) {
    return 0;
  }

  for(i=1; i<=4; i++) {
    char *s=str+pmatch[i+1].rm_so;
    char *e=str+pmatch[i+1].rm_eo;
    int x=0;
    if (*s=='0' && e-s!=1) {
      if (!errmsg) errmsg="Invalid IP format";
    }
    while(s<e) {
      x=x*10+ (*s-'0');
      s++;
    }
    if (x>255) {
      if (!errmsg) errmsg= "Invalid IP address";
    }
    ip[i-1]=x;
  }

  if (ip[0]>=224) {
    if (!errmsg) errmsg = "Multicast IP address";
  }
  if (query) {
    int len = pmatch[1].rm_eo-pmatch[1].rm_so;
    strncpy(query, str+pmatch[1].rm_so, len);
    query[len]=0;
  }
  if (errmsg) {
    if (txt) strcpy(txt, errmsg);
    if (a)   strcpy(a, "127.0.0.2");
    return -1;
  } else {
    if (txt) txt[0]=0;
    if (a) a[0]=0;
  }
  return 1;
}

/* entry found */
#define BLFLAG_FOUND     0x01
/* entry not found after a full research */
#define BLFLAG_NOTFOUND  0x02
/* entry not found on a level 1 search */
#define BLFLAG_LEVEL1    0x04

/*
 * check in RHSBL blacklist type
 */

static struct blresult *check_rhsbl(struct fstate *state,
				    struct slist_elt *rbl_e ) {
  struct blresult *res=NULL;
  char *domain;
  char buf[255];
  int exists;

  DEBUG("check RHSBL"); 

  switch(rbl_e->flags) {
  case BLFLAG_FOUND:
    return (struct blresult *)rbl_e->extra;
  case BLFLAG_NOTFOUND:
    return NULL;
  }

  if (state->sender[0] == 0)
    return NULL;

  domain = strrchr(state->sender, '@');
  if (domain == NULL)
    return NULL;
  domain ++ ; /* skip @ */

  if ((res = malloc(sizeof(struct blresult)))==NULL)
    return NULL;

  snprintf(buf, sizeof(buf), "%s.%s.", domain, rbl_e->value);
  exists=dns_exists(state, buf, res->a, res->txt);
  DEBUG("exists=%d buf=%s a=%s", exists, buf, res->a);
  if (exists < 0) { /* lookup error */
    free(res);
    return NULL; 
  }
  if (exists == 0) {
    free(res);
    rbl_e->flags = BLFLAG_NOTFOUND;
    return NULL;
  }

  if (rbl_e->extra)
    free(rbl_e->extra);
  strncpy(res->query, domain, sizeof(res->query)-1);
  res->query[sizeof(res->query)-1]=0;
  rbl_e->extra = res;
  rbl_e->flags = BLFLAG_FOUND;
  res->rbl   = rbl_e->key;
  res->proof = 0;
  res->header = NULL;
  res->domain = rbl_e->value;
  return res;
}

/*
 *  check if a Received is in a RBL.
 *  return 1 if yes
 *  return 0 if no
 */
static struct blresult *check_rbl(struct fstate *state,
				  struct slist_elt *rbl_e,
				  int level1) {
  struct headers *hl;
  int proof = 0;
  int minproof = 0;
  struct blresult *res=NULL;

  DEBUG("check RBL"); 

  if (rbl_e->flags) {
    struct blresult *extra = (struct blresult *)rbl_e->extra;
    DEBUG("(cached)");
    if (level1) {
      return extra && extra->proof>1 ? NULL : extra;
    } else {
      if (extra)
	return extra;
      if (rbl_e->flags == BLFLAG_NOTFOUND)
	return NULL ;
      minproof = 1;  /* search again */
    }
  } 
  if ((res = malloc(sizeof(struct blresult)))==NULL)
      return NULL;
  rbl_e->flags = level1 ? BLFLAG_LEVEL1 : BLFLAG_NOTFOUND  ;
  
  hl=state->hdr_in_received;

  while(hl) {
    int ip[4];

    if (hl->type == HEADER_RECEIVED) {
      int invalid = 0;
      proof++;
      DEBUG("%s", hl->header);
      if (proof <= minproof) {
	DEBUG("skiped");
	hl=hl->next;
	continue;
      }
      if (rbl_e->value[0]==0) {
	int ret = get_addr_from_received(hl->header, ip, 
					 res->query, res->a, res->txt);
	DEBUG("ret=%d", ret);
	if (ret < 0) invalid = 1;
      } else {
	int ret = get_addr_from_received(hl->header, ip,
					 res->query, NULL, NULL);
	if (ret==1) {
	  char buf[255];
	  int exists;
	  snprintf(buf, sizeof(buf), "%d.%d.%d.%d.%s.",
		     ip[3], ip[2], ip[1], ip[0], rbl_e->value);
	  exists=dns_exists(state, buf, res->a, res->txt);
	  DEBUG("ret=%d exists=%d buf=%s a=%s", ret, exists, buf, res->a);
	  if (exists > 0) invalid = 1;
	}
      }
      DEBUG("Check(%s) %s => %d (%s %s)",
	    res->query, rbl_e->key, invalid, res->a, res->txt);

      if (invalid) {
	if (rbl_e->extra)
	  free(rbl_e->extra);
	rbl_e->extra = res;
	rbl_e->flags = BLFLAG_FOUND;
	res->rbl     = rbl_e->key;
	res->proof   = proof;
	res->header  = hl;
	res->domain  = rbl_e->value;
	return res;
      }
      if (level1)
	break;
    }
    hl = hl->next;
  }
  free(res);
  return NULL;
  
}

/*
 * check in blacklist 
 */


struct blresult *checkbl(struct fstate *state, char *rbl) {
  struct slist_elt *rbl_e; 
  int len;
  int level1;

  DEBUG("check in %s", rbl); 
  
  rbl_e = slist_search(&(state->rhsbl_server), rbl);
  if (rbl_e)
    return check_rhsbl(state, rbl_e);

  len = strlen(rbl);
  if (len>0 && rbl[len-1]=='*') {
    rbl[len-1]=0;
    level1=0;
  } else {
    level1=1;
  }
  rbl_e = slist_search(&(state->rbl_server), rbl);

  if (rbl_e) 
    return check_rbl(state, rbl_e, level1);

  return NULL;
}
